#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

from typing import Optional

from s3torchconnectorclient._mountpoint_s3_client import (
    MockMountpointS3Client,
    MountpointS3Client,
)

from . import S3Client
from .._user_agent import UserAgent
from .s3client_config import S3ClientConfig

"""
_mock_s3client.py
    Internal client wrapper mock class for unit testing.
"""


class MockS3Client(S3Client):
    def __init__(
        self,
        region: str,
        bucket: str,
        user_agent: Optional[UserAgent] = None,
        s3client_config: Optional[S3ClientConfig] = None,
    ):
        super().__init__(
            region,
            user_agent=user_agent,
            s3client_config=s3client_config,
        )
        self._mock_client = MockMountpointS3Client(
            region,
            bucket,
            throughput_target_gbps=self.s3client_config.throughput_target_gbps,
            part_size=self.s3client_config.part_size,
            user_agent_prefix=self.user_agent_prefix,
            unsigned=self.s3client_config.unsigned,
            force_path_style=self.s3client_config.force_path_style,
            max_attempts=self.s3client_config.max_attempts,
        )

    def add_object(self, key: str, data: bytes) -> None:
        self._mock_client.add_object(key, data)

    def remove_object(self, key: str) -> None:
        self._mock_client.remove_object(key)

    def _client_builder(self) -> MountpointS3Client:
        return self._mock_client.create_mocked_client()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

#  Get a nice error message if lightning isn't available.
import lightning

from .s3_lightning_checkpoint import S3LightningCheckpoint

__all__ = [
    "S3LightningCheckpoint",
]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import logging
import os
import gc
import threading
import traceback
import weakref
from functools import partial
from typing import Optional, Any, List

from s3torchconnector import S3Reader, S3Writer, S3ReaderConstructor
from s3torchconnector.s3reader.protocol import S3ReaderConstructorProtocol
from .s3client_config import S3ClientConfig

from s3torchconnectorclient._mountpoint_s3_client import (
    MountpointS3Client,
    ObjectInfo,
    HeadObjectResult,
    ListObjectStream,
    GetObjectStream,
    join_all_managed_threads,
)

from s3torchconnector._user_agent import UserAgent

"""
_s3client.py
    Internal client wrapper class on top of S3 client implementation
    with multi-process support.
"""


log = logging.getLogger(__name__)


def _identity(obj: Any) -> Any:
    return obj


_client_lock = threading.Lock()
_active_clients: "weakref.WeakSet[Any]" = weakref.WeakSet()


def _before_fork_handler():
    # Handler that cleans up CRT resources before fork operations.
    try:
        # Get all instances of S3Client that exist in the current process
        clients_list = _get_active_s3clients()
        if not clients_list:
            return

        for client in clients_list:
            if client._native_client is not None:
                # Release the client before fork as it's not fork-safe
                client._native_client = None
        # Clear the list of active clients. We will re-populate it after fork with only active clients.
        _reset_active_s3clients()
        gc.collect()
        # Wait for native background threads to complete joining (0.5 sec timeout)
        join_all_managed_threads(0.5)
    except Exception as e:
        print(
            "Warning: Failed to properly clean up native background threads before fork. "
            f"Error: {e}\n"
            "Your subprocess may crash or hang. To prevent this:\n"
            "1. Ensure no active S3 client usage during fork operations\n"
            "2. Use multiprocessing with 'spawn' or 'forkserver' start method instead"
        )
        traceback.print_exc()


def _after_fork_handler():
    """Handler that initializes the list of active clients after fork in child process."""
    _reset_active_s3clients()


# register the handler to release the S3 client and wait for background threads to join before fork happens
# As fork will not inherit any background threads. Wait for them to join to avoid crashes or hangs.
os.register_at_fork(before=_before_fork_handler, after_in_child=_after_fork_handler)


class S3Client:
    def __init__(
        self,
        region: str,
        *,
        endpoint: Optional[str] = None,
        user_agent: Optional[UserAgent] = None,
        s3client_config: Optional[S3ClientConfig] = None,
    ):
        self._region = region
        self._endpoint = endpoint
        user_agent = user_agent or UserAgent()
        self._user_agent_prefix = user_agent.prefix
        self._s3client_config = s3client_config or S3ClientConfig()
        self._client_pid: Optional[int] = None
        self._native_client: Optional[MountpointS3Client] = None

    @property
    def _client(self) -> MountpointS3Client:
        if (
            self._client_pid is None
            or self._client_pid != os.getpid()
            or self._native_client is None
        ):
            # Acquire the lock to ensure thread-safety when creating the client.
            with _client_lock:
                if (
                    self._client_pid is None
                    or self._client_pid != os.getpid()
                    or self._native_client is None
                ):
                    # This double-check ensures that the client is only created once.
                    self._native_client = self._client_builder()
                    self._client_pid = os.getpid()
                    # Track the client in the list of active clients.
                    _active_clients.add(self)

        assert self._native_client is not None
        return self._native_client

    @property
    def region(self) -> str:
        return self._region

    @property
    def s3client_config(self) -> S3ClientConfig:
        return self._s3client_config

    @property
    def user_agent_prefix(self) -> str:
        return self._user_agent_prefix

    def _client_builder(self) -> MountpointS3Client:
        return MountpointS3Client(
            region=self._region,
            endpoint=self._endpoint,
            profile=self._s3client_config.profile,
            user_agent_prefix=self._user_agent_prefix,
            throughput_target_gbps=self._s3client_config.throughput_target_gbps,
            part_size=self._s3client_config.part_size,
            unsigned=self._s3client_config.unsigned,
            force_path_style=self._s3client_config.force_path_style,
            max_attempts=self._s3client_config.max_attempts,
        )

    def get_object(
        self,
        bucket: str,
        key: str,
        *,
        object_info: Optional[ObjectInfo] = None,
        reader_constructor: Optional[S3ReaderConstructorProtocol] = None,
    ) -> S3Reader:
        log.debug(f"GetObject s3://{bucket}/{key}, {object_info is None=}")
        if object_info is None:
            get_object_info = partial(self.head_object, bucket, key)
        else:
            get_object_info = partial(_identity, object_info)

        reader_constructor = reader_constructor or S3ReaderConstructor.default()

        return reader_constructor(
            bucket=bucket,
            key=key,
            get_object_info=get_object_info,
            get_stream=partial(self._get_object_stream, bucket, key),
        )

    def _get_object_stream(
        self,
        bucket: str,
        key: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> GetObjectStream:
        return self._client.get_object(bucket, key, start, end)

    def put_object(
        self, bucket: str, key: str, storage_class: Optional[str] = None
    ) -> S3Writer:
        log.debug(f"PutObject s3://{bucket}/{key}")
        return S3Writer(self._client.put_object(bucket, key, storage_class))

    # TODO: Probably need a ListObjectResult on dataset side
    def list_objects(
        self, bucket: str, prefix: str = "", delimiter: str = "", max_keys: int = 1000
    ) -> ListObjectStream:
        log.debug(f"ListObjects s3://{bucket}/{prefix}")
        return self._client.list_objects(bucket, prefix, delimiter, max_keys)

    # TODO: We need ObjectInfo on dataset side
    def head_object(self, bucket: str, key: str) -> HeadObjectResult:
        log.debug(f"HeadObject s3://{bucket}/{key}")
        return self._client.head_object(bucket, key)

    def delete_object(self, bucket: str, key: str) -> None:
        log.debug(f"DeleteObject s3://{bucket}/{key}")
        self._client.delete_object(bucket, key)

    def copy_object(
        self, src_bucket: str, src_key: str, dst_bucket: str, dst_key: str
    ) -> None:
        log.debug(
            f"CopyObject s3://{src_bucket}/{src_key} to s3://{dst_bucket}/{dst_key}"
        )
        return self._client.copy_object(src_bucket, src_key, dst_bucket, dst_key)


def _get_active_s3clients() -> List[S3Client]:
    """
    Returns a copy of list of all active S3 clients.
    Pay attention, it grabs a lock on _client_lock.

    Returns:
        List[S3Client]: A list of active S3 client instances.
    """
    with _client_lock:
        return list(_active_clients)


def _reset_active_s3clients():
    """
    Resets the list of active S3 clients.
    Pay attention, it grabs a lock on _client_lock.
    """
    global _active_clients
    with _client_lock:
        _active_clients = weakref.WeakSet()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import dataclasses
import io
import logging
import os
import urllib.parse
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Union, Optional, List

import torch
from s3torchconnectorclient._mountpoint_s3_client import S3Exception
from tenacity import (
    retry,
    stop_after_attempt,
    retry_if_exception_type,
    before_sleep_log,
    after_log,
    wait_random_exponential,
)
from torch.distributed.checkpoint.filesystem import (
    FileSystemReader,
    FileSystemWriter,
    FileSystemBase,
)
from torch.distributed.checkpoint.planner import SavePlan, LoadPlan


from s3torchconnector._s3client import S3Client
from s3torchconnector._s3dataset_common import parse_s3_uri
from ..s3reader import (
    S3ReaderConstructor,
    S3ReaderConstructorProtocol,
    DCPS3ReaderConstructorProtocol,
)
from .. import S3ClientConfig
from .s3_prefix_strategy import S3PrefixStrategyBase, DefaultPrefixStrategy
from .._user_agent import UserAgent

logger = logging.getLogger(__name__)


class S3FileSystem(FileSystemBase):
    """S3-based implementation of PyTorch's FileSystemBase for distributed checkpointing."""

    def __init__(
        self,
        region: str,
        s3_client: Optional[S3Client] = None,
        s3client_config: Optional[S3ClientConfig] = None,
        reader_constructor: Optional[S3ReaderConstructorProtocol] = None,
    ) -> None:
        """
        Initialize S3FileSystem.

        Args:
            region (str): The AWS region for S3.
            s3_client (Optional[S3Client]): Optional S3Client instance.
            s3client_config (Optional[S3ClientConfig]): Optional S3ClientConfig with parameters for S3 client.
            reader_constructor (Optional[S3ReaderConstructorProtocol]): Optional partial(S3Reader) created using S3ReaderConstructor
                e.g. S3ReaderConstructor.sequential() or S3ReaderConstructor.range_based()
        """
        self._path: Union[str, os.PathLike] = ""
        self._reader_constructor = reader_constructor or S3ReaderConstructor.default()

        # Get reader type string for user agent
        reader_type_string = S3ReaderConstructor.get_reader_type_string(
            self._reader_constructor
        )
        user_agent = UserAgent(
            ["dcp", torch.__version__, f"md/reader_type#{reader_type_string}"]
        )

        self._client = (
            S3Client(
                region=region, user_agent=user_agent, s3client_config=s3client_config
            )
            if s3_client is None
            else s3_client
        )

    @contextmanager
    def create_stream(
        self, path: Union[str, os.PathLike], mode: str
    ) -> Generator[io.IOBase, None, None]:
        """
        Create a stream for reading or writing to S3.

        Args:
            path (Union[str, os.PathLike]): The S3 path to read or write.
            mode (str): The mode for the stream. Supports 'rb' for read mode and 'wb' for write mode.

        Yields:
            io.BufferedIOBase: A stream for reading or writing to S3.

        Raises:
            ValueError: If the mode is not 'rb' or 'wb'.
        """
        path_str = _path_or_str_to_str(path)
        bucket, key = parse_s3_uri(path_str)

        if mode == "wb":  # write mode
            logger.debug("create_stream writable for %s", path_str)
            with self._client.put_object(bucket, key) as stream:
                yield stream
        elif mode == "rb":  # read mode
            logger.debug("create_stream readable for %s", path_str)
            with self._client.get_object(
                bucket, key, reader_constructor=self._reader_constructor
            ) as stream:
                yield stream
        else:
            raise ValueError(
                f"Invalid {mode=} mode argument: create_stream only supports rb (read mode) & wb (write mode)"
            )

    def concat_path(self, path: Union[str, os.PathLike], suffix: str) -> str:
        """
        Concatenate a suffix to the given path.

        Args:
            path (Union[str, os.PathLike]): The base path.
            suffix (str): The suffix to concatenate.

        Returns:
            str: The concatenated path.
        """
        logger.debug("concat paths %s and %s", path, suffix)
        path_str = os.fspath(path)
        result = os.path.join(path_str, suffix)
        return result

    def init_path(self, path: Union[str, os.PathLike]) -> Union[str, os.PathLike]:
        """
        Initialize the path for the filesystem.

        Args:
            path (Union[str, os.PathLike]): The path to initialize.

        Returns:
            Union[str, os.PathLike]: The initialized path.
        """
        logger.debug("init_path for %s", path)
        self._path = path
        return self._path

    def rename(
        self, old_path: Union[str, os.PathLike], new_path: Union[str, os.PathLike]
    ) -> None:
        """Rename an object in S3.

        This is emulated by copying it to a new path and deleting the old path. The deletion part is retried (see also
        :func:`S3FileSystem._delete_with_retry`).

        Args:
            old_path (Union[str, os.PathLike]): The current path of the object.
            new_path (Union[str, os.PathLike]): The new path for the object.

        Raises:
            ValueError: If the old and new paths point to different buckets.
            S3Exception: If there is an error with the S3 client.
        """
        logger.debug("rename %s to %s", old_path, new_path)

        old_path_str = _path_or_str_to_str(old_path)
        new_path_str = _path_or_str_to_str(new_path)

        old_bucket, old_key = parse_s3_uri(old_path_str)
        escaped_old_key = self._escape_path(old_key)
        logger.debug("rename: escaped version of the source key: %s", escaped_old_key)
        new_bucket, new_key = parse_s3_uri(new_path_str)

        if old_bucket != new_bucket:
            raise ValueError(
                f"Source and destination buckets cannot be different (rename does not support cross-buckets operations)"
            )

        self._client.copy_object(
            src_bucket=old_bucket,
            src_key=escaped_old_key,
            dst_bucket=new_bucket,
            dst_key=new_key,
        )
        logger.debug("rename: copied %s to %s successfully", old_path_str, new_path_str)
        self._delete_with_retry(old_bucket, old_key)
        logger.debug("rename: s3://%s/%s successfully", old_bucket, old_key)

    def mkdir(self, path: Union[str, os.PathLike]) -> None:
        """No-op method for creating directories in S3 (not needed)."""
        pass

    def exists(self, path: Union[str, os.PathLike]) -> bool:
        logger.debug("exists %s", path)

        path_str = _path_or_str_to_str(path)
        bucket, key = parse_s3_uri(path_str)
        try:
            self._client.head_object(bucket, key)
        except S3Exception as e:
            if str(e) != "Service error: The object was not found":
                raise
            return False
        return True

    def rm_file(self, path: Union[str, os.PathLike]) -> None:
        logger.debug("remove %s", path)

        path_str = _path_or_str_to_str(path)
        bucket, key = parse_s3_uri(path_str)
        try:
            self._client.delete_object(bucket, key)
        except S3Exception:
            logger.exception("Failed to remove object from S3")

    @classmethod
    def validate_checkpoint_id(cls, checkpoint_id: Union[str, os.PathLike]) -> bool:
        logger.debug("validate_checkpoint_id for %s", checkpoint_id)

        if isinstance(checkpoint_id, Path):
            return True

        try:
            parse_s3_uri(_path_or_str_to_str(checkpoint_id))
        except ValueError:
            return False
        return True

    @retry(
        retry=retry_if_exception_type(S3Exception),
        stop=stop_after_attempt(3),
        wait=wait_random_exponential(multiplier=1, max=5),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        after=after_log(logger, logging.ERROR),
        reraise=True,
    )
    def _delete_with_retry(self, bucket_name: str, old_key: str):
        """Wrapper around :func:`S3Client.delete_object` to retry the deletion.

        Will retry a maximum of 3 times, only for `S3Exception`s, and wait between retries. It will reraise the caught
        exception too, and logs retries and final error, if any."""
        self._client.delete_object(bucket_name, old_key)

    @staticmethod
    def _escape_path(string):
        """URL-encodes path segments while preserving '/' separators using urllib.parse.quote().

        Args:
            string (str): URL path string to escape

        Returns:
            str: Path string with each segment percent-encoded, separators preserved
        """
        if not string:
            return string
        parts = []
        for part in string.split("/"):
            parts.append(urllib.parse.quote(part, safe=""))
        return "/".join(parts)


@dataclass
class StorageMetadata:
    """Metadata for S3 storage prefix."""

    prefix: str


class S3StorageWriter(FileSystemWriter):
    """S3 implementation of PyTorch's FileSystemWriter for distributed checkpoints."""

    def __init__(
        self,
        region: str,
        path: str,
        s3client_config: Optional[S3ClientConfig] = None,
        prefix_strategy: Optional[S3PrefixStrategyBase] = None,
        thread_count: int = 1,
        **kwargs,
    ) -> None:
        """
        Initialize an S3 writer for distributed checkpointing.

        Args:
            region (str): The AWS region for S3.
            path (str): The S3 URI to write checkpoints to.
            s3client_config (Optional[S3ClientConfig]): Optional S3ClientConfig with parameters for S3 client.
            prefix_strategy (Optional[S3PrefixStrategyBase]): Optional strategy for generating S3 prefixes to
                optimize checkpoint organization and prevent throttling.
            thread_count (int): Number of IO threads to use to write. Defaults to 1 (Pytorch Default)
            kwargs (dict): Keyword arguments to pass to the parent :class:`FileSystemWriter`.
        """
        super().__init__(
            path=path,
            sync_files=False,  # FIXME: setting this to True makes the run to fail (L#333: `os.fsync(stream.fileno())`)
            thread_count=thread_count,
            **kwargs,
        )
        self.fs = S3FileSystem(region, s3client_config=s3client_config)  # type: ignore
        self.path = self.fs.init_path(path)
        self.prefix_strategy = prefix_strategy or DefaultPrefixStrategy()

    def prepare_global_plan(self, plans: List[SavePlan]) -> List[SavePlan]:
        """
        Prepare save plans with S3-specific storage metadata.

        Args:
            plans: List of save plans to be processed.

        Returns:
            Modified save plans with S3 storage metadata.
        """
        return [
            dataclasses.replace(
                plan, storage_data=StorageMetadata(self.prefix_strategy(idx))
            )
            for idx, plan in enumerate(plans)
        ]

    @classmethod
    def validate_checkpoint_id(cls, checkpoint_id: Union[str, os.PathLike]) -> bool:
        return S3FileSystem.validate_checkpoint_id(checkpoint_id)


class S3StorageReader(FileSystemReader):
    """S3 implementation of PyTorch's FileSystemReader with configurable reader strategies.

    By default, uses DCPOptimizedS3Reader for improved checkpoint loading performance.
    For unsupported or non-DCP access patterns, please use the generic reader:
        storage_reader = S3StorageReader(
            region, path,
            reader_constructor=S3ReaderConstructor.default()
        )
    """

    def __init__(
        self,
        region: str,
        path: Union[str, os.PathLike],
        s3client_config: Optional[S3ClientConfig] = None,
        reader_constructor: Optional[
            Union[S3ReaderConstructorProtocol, DCPS3ReaderConstructorProtocol]
        ] = None,
    ) -> None:
        """
        Initialize an S3 reader for distributed checkpointing.

        Args:
            region (str): The AWS region for S3.
            path (Union[str, os.PathLike]): The S3 path to read checkpoints from.
            s3client_config (Optional[S3ClientConfig]): Optional S3ClientConfig with parameters for S3 client.
            reader_constructor (Optional[S3ReaderConstructorProtocol]): Reader constructor created using
                S3ReaderConstructor. Defaults to ``S3ReaderConstructor.dcp_optimized()`` for best performance.
                Use ``S3ReaderConstructor.sequential()`` for unsupported/non-DCP access patterns.
        """
        super().__init__(path)
        self._reader_constructor = (
            reader_constructor or S3ReaderConstructor.dcp_optimized()
        )
        self.fs: S3FileSystem = S3FileSystem(  # type: ignore[assignment] # since we overrode self.fs: FileSystem
            region,
            s3client_config=s3client_config,
            reader_constructor=self._reader_constructor,
        )
        self.path = self.fs.init_path(path)
        self.sync_files = False

    @classmethod
    def validate_checkpoint_id(cls, checkpoint_id: Union[str, os.PathLike]) -> bool:
        return S3FileSystem.validate_checkpoint_id(checkpoint_id)

    def prepare_local_plan(self, plan: LoadPlan) -> LoadPlan:
        """
        Performs two key optimizations:

            1. **Load Ordering**: Sorts load items by storage offset to enable sequential access

            2. **Range Injection**: Provides byte range metadata to DCP reader constructors to enable
            usage of DCPOptimizedS3Reader for range-based streams and range coalescing

        Args:
            plan (LoadPlan): The load plan from PyTorch DCP.

        Returns:
            LoadPlan: The same plan with items sorted by storage offset.

        Note:
            Both optimizations are required for DCPOptimizedS3Reader.
        """
        # Sort items in plan based on their offset in checkpoints shards
        plan.items.sort(key=lambda item: self.storage_data[item.storage_index].offset)

        # Inject ranges if using DCP optimized reader constructor
        if isinstance(self._reader_constructor, DCPS3ReaderConstructorProtocol):
            self._reader_constructor.set_item_ranges_by_file(
                plan.items, self.storage_data, self.path
            )

        return plan


def _path_or_str_to_str(path: Union[str, os.PathLike]) -> str:
    return path if isinstance(path, str) else str(path)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD
from abc import ABC, abstractmethod
from typing import List, Optional

import torch.distributed as dist


class S3PrefixStrategyBase(ABC):
    """Base class for S3 prefix generation strategies."""

    def __init__(self):
        pass

    def __call__(self, rank: int) -> str:
        """Generate prefix for given rank."""
        return self.generate_prefix(rank)

    @abstractmethod
    def generate_prefix(self, rank: int) -> str:
        """Generate storage prefix for the given rank."""
        pass


class DefaultPrefixStrategy(S3PrefixStrategyBase):
    """Default strategy for generating S3 prefixes."""

    def generate_prefix(self, rank: int) -> str:
        """Generate simple rank-based name without prefix."""
        return f"__{rank}_"


class NumericPrefixStrategy(S3PrefixStrategyBase):
    """Base class for numeric prefix generation strategies."""

    def __init__(
        self,
        base: int,
        epoch_num: Optional[int] = None,
        min_prefix_length: int = 10,
        prefix_count: Optional[int] = None,
    ):
        """
            Initialize numeric prefix strategy.

        Args:
            base (int): The numeric base for the prefix (e.g., 2 for binary, 16 for hex).
            epoch_num (int, optional): Epoch number for checkpoint ordering. If None,
                no epoch information will be included in the prefix. Defaults to None.
            min_prefix_length (int): Minimum length of the generated prefix. Prefix will be
                padded with trailing zeros if necessary. Must be positive. Defaults to 10.
            prefix_count (int, optional): Number of unique prefixes to generate. If not provided,
                world size will be used as default value. Defaults to None.

        Raises:
            ValueError: If epoch_num, min_prefix_length, or prefix_count are invalid.
        """
        if min_prefix_length < 1:
            raise ValueError(
                f"Minimum prefix length must be positive, got {min_prefix_length}"
            )

        if epoch_num is not None and not isinstance(epoch_num, int):
            raise ValueError(
                f"Epoch number must be None or an integer, got {epoch_num}"
            )

        if prefix_count is not None and (
            not isinstance(prefix_count, int) or prefix_count < 1
        ):
            raise ValueError(
                f"Prefix count must be a positive integer, got {prefix_count}"
            )

        super().__init__()
        self.base = base
        self.epoch_num = epoch_num
        self.min_prefix_len = min_prefix_length

        self.prefix_count = 1
        if prefix_count is not None:
            self.prefix_count = prefix_count
        elif dist.is_initialized():
            self.prefix_count = dist.get_world_size()

        self.prefix_map = self._generate_prefix_map()

    def generate_prefix(self, rank: int) -> str:
        """
        Generate numeric-based prefix with optional epoch number.

        Args:
            rank: Process rank in the distributed environment.

        Returns:
            Prefix string in format: <pattern>/epoch_<num>/__<rank>_
            or <pattern>/__<rank>_ if no epoch number is provided.
        """
        epoch_suffix = f"epoch_{self.epoch_num}/" if self.epoch_num is not None else ""
        return f"{self.prefix_map[rank % len(self.prefix_map)]}/{epoch_suffix}__{rank}_"

    def _generate_prefix_map(self) -> List[str]:
        """Generate mapping of ranks to numeric-based prefixes."""
        minimum_required_length = self._calculate_prefix_length()
        adjusted_prefix_length = max(minimum_required_length, self.min_prefix_len)

        all_prefixes = [
            self._format_number(i, adjusted_prefix_length)[::-1]
            for i in range(self.prefix_count)
        ]

        return all_prefixes

    def _calculate_prefix_length(self) -> int:
        """Calculate minimum prefix length needed for unique combinations."""
        prefix_length = 1
        size = self.base
        while size < self.prefix_count:
            prefix_length += 1
            size *= self.base
        return prefix_length

    @abstractmethod
    def _format_number(self, number: int, length: int) -> str:
        """Format a number to the appropriate base representation."""
        pass


class BinaryPrefixStrategy(NumericPrefixStrategy):
    """Binary (Base2) prefix generation strategy using only 0 and 1."""

    def __init__(
        self,
        epoch_num: Optional[int] = None,
        min_prefix_length: int = 10,
        prefix_count: Optional[int] = None,
    ):
        super().__init__(
            base=2,
            epoch_num=epoch_num,
            min_prefix_length=min_prefix_length,
            prefix_count=prefix_count,
        )

    def _format_number(self, number: int, length: int) -> str:
        return format(number, f"0{length}b")


class HexPrefixStrategy(NumericPrefixStrategy):
    """Hexadecimal-based prefix generation strategy."""

    def __init__(
        self,
        epoch_num: Optional[int] = None,
        min_prefix_length: int = 10,
        prefix_count: Optional[int] = None,
    ):
        super().__init__(
            base=16,
            epoch_num=epoch_num,
            min_prefix_length=min_prefix_length,
            prefix_count=prefix_count,
        )

    def _format_number(self, number: int, length: int) -> str:
        return format(number, f"0{length}x")


class RoundRobinPrefixStrategy(S3PrefixStrategyBase):
    """Strategy that distributes ranks across user-provided prefixes in round-robin fashion."""

    def __init__(self, user_prefixes: List[str], epoch_num: Optional[int] = None):
        """
        Initialize round-robin prefix strategy.

        Args:
            user_prefixes: List of prefixes to distribute ranks across.
                          Must not be empty.
            epoch_num: Epoch number for checkpoint ordering.

        Raises:
            ValueError: If user_prefixes is empty.
        """
        super().__init__()
        if not user_prefixes:
            raise ValueError("user_prefixes must not be empty")

        self.user_prefixes = user_prefixes
        self.epoch_num = epoch_num

    def generate_prefix(self, rank: int) -> str:
        """
        Generate prefix for given rank using round-robin distribution.

        Args:
            rank: Process rank in the distributed environment.

        Returns:
            Prefix string in format: <user_prefix>/epoch_<num>/__<rank>_
            or <user_prefix>/__<rank>_ if no epoch number is provided.
        """
        epoch_suffix = f"epoch_{self.epoch_num}/" if self.epoch_num is not None else ""
        return f"{self.user_prefixes[rank % len(self.user_prefixes)]}/{epoch_suffix}__{rank}_"



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

from .s3_file_system import S3FileSystem, S3StorageReader, S3StorageWriter
from .s3_prefix_strategy import (
    S3PrefixStrategyBase,
    DefaultPrefixStrategy,
    NumericPrefixStrategy,
    BinaryPrefixStrategy,
    HexPrefixStrategy,
)

__all__ = [
    "S3FileSystem",
    "S3StorageReader",
    "S3StorageWriter",
    "S3PrefixStrategyBase",
    "DefaultPrefixStrategy",
    "NumericPrefixStrategy",
    "BinaryPrefixStrategy",
    "HexPrefixStrategy",
]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import io
from functools import cached_property
from io import SEEK_CUR, SEEK_END, SEEK_SET
from typing import Callable, Optional, Iterator, Union

from s3torchconnectorclient._mountpoint_s3_client import (
    ObjectInfo,
    GetObjectStream,
    HeadObjectResult,
)
from .s3reader import S3Reader


class SequentialS3Reader(S3Reader):
    """Sequential S3 reader implementation

    Maintains an internal buffer for efficient sequential reads and repeated access.
    Optimal for most use cases, including full object reads.
    """

    def __init__(
        self,
        bucket: str,
        key: str,
        get_object_info: Callable[[], Union[ObjectInfo, HeadObjectResult]],
        get_stream: Callable[[], GetObjectStream],
    ):
        if not bucket:
            raise ValueError("Bucket should be specified")
        self._bucket = bucket
        self._key = key
        self._get_object_info = get_object_info
        self._get_stream = get_stream
        self._stream: Optional[Iterator[bytes]] = None
        self._buffer = io.BytesIO()
        self._size: Optional[int] = None
        # Invariant: _position == _buffer._tell() unless _position_at_end()
        self._position = 0

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def key(self) -> str:
        return self._key

    @cached_property
    def _object_info(self):
        return self._get_object_info()

    def prefetch(self) -> None:
        """Start fetching data from S3.

        Raises:
            S3Exception: An error occurred accessing S3.
        """

        if self._stream is None:
            self._stream = self._get_stream()

    def readinto(self, buf) -> int:
        """Read up to len(buf) bytes into a pre-allocated, writable bytes-like object buf.
        Return the number of bytes read. If no bytes are available, zero is returned.

        Args:
            buf : writable bytes-like object

        Returns:
            int : numer of bytes read or zero, if no bytes available
        """
        buf_size = len(buf)
        if self._position_at_end() or buf_size == 0:
            # If no bytes are available or no place to write data, zero should be returned
            return 0

        self.prefetch()
        assert self._stream is not None

        cur_pos = self._position
        # preload enough bytes in buffer
        self.seek(buf_size, SEEK_CUR)
        # restore position, before starting to write into buf
        self._buffer.seek(cur_pos)
        size = self._buffer.readinto(buf)
        self._position = self._buffer.tell()

        return size

    def read(self, size: Optional[int] = None) -> bytes:
        """Read up to size bytes from the object and return them.

        If size is zero or positive, read that many bytes from S3, or until the end of the object.
        If size is None or negative, read the entire file.

        Args:
            size (int | None): how many bytes to read.

        Returns:
            bytes: Bytes read from S3 Object

        Raises:
            S3Exception: An error occurred accessing S3.
        """

        if size is not None and not isinstance(size, int):
            raise TypeError(f"argument should be integer or None, not {type(size)!r}")
        if self._position_at_end():
            # Invariant: if we're at EOF, it doesn't matter what `size` is, we'll always return no data and have no
            # side effect.
            return b""

        self.prefetch()
        assert self._stream is not None
        cur_pos = self._position
        if size is None or size < 0:
            # Special case read() all to use O(n) algorithm
            self._buffer.seek(0, SEEK_END)
            for batch in self._stream:
                self._buffer.write(batch)

            # Once we've emptied the buffer, we'll always be at EOF!
            self._size = self._buffer.tell()
        else:
            self.seek(size, SEEK_CUR)

        self._buffer.seek(cur_pos)
        data = self._buffer.read(size)
        self._position = self._buffer.tell()
        return data

    def seek(self, offset: int, whence: int = SEEK_SET, /) -> int:
        """Change the stream position to the given byte offset, interpreted relative to whence.

        When seeking beyond the end of the file, always stay at EOF.
        Seeking before the start of the file results in a ValueError.

        Args:
            offset (int): How many bytes to seek relative to whence.
            whence (int): One of SEEK_SET, SEEK_CUR, and SEEK_END. Default: SEEK_SET

        Returns:
            int: Current position of the stream

        Raises:
            S3Exception: An error occurred accessing S3.

        """
        if not isinstance(offset, int):
            raise TypeError(f"integer argument expected, got {type(offset)!r}")
        if whence == SEEK_END:
            if offset >= 0:
                self._position = self._get_size()
                return self._position
            offset += self._get_size()
        elif whence == SEEK_CUR:
            if self._position_at_end() and offset >= 0:
                return self._position
            offset += self._position
        elif whence == SEEK_SET:
            pass
        elif isinstance(whence, int):
            raise ValueError("Seek must be passed SEEK_CUR, SEEK_SET, or SEEK_END")
        else:
            raise TypeError(f"integer argument expected, got {type(whence)!r}")

        if offset < 0:
            raise ValueError(f"negative seek value {offset}")

        if offset > self._buffer_size():
            self._prefetch_to_offset(offset)

        if self._size is not None:
            offset = min(offset, self._size)

        self._position = self._buffer.seek(offset)
        return self._position

    def _prefetch_to_offset(self, offset: int) -> None:
        self.prefetch()
        assert self._stream is not None
        buf_size = self._buffer.seek(0, SEEK_END)
        try:
            while offset > buf_size:
                buf_size += self._buffer.write(next(self._stream))
        except StopIteration:
            self._size = self._buffer.tell()

    def _get_size(self) -> int:
        if self._size is None:
            self._size = self._object_info.size
        return self._size

    def _position_at_end(self) -> bool:
        # Code calling this must only be used for optimisation purposes.
        if self._size is None:
            # We can never be special cased to EOF if we never saw how long it is.
            # If we _are_ at EOF, we'll just not take the early exits.
            return False
        return self._position == self._size

    def _buffer_size(self) -> int:
        cur_pos = self._buffer.tell()
        self._buffer.seek(0, SEEK_END)
        buffer_size = self._buffer.tell()
        self._buffer.seek(cur_pos)
        return buffer_size

    def tell(self) -> int:
        """
        Returns:
              int: Current stream position.
        """
        return self._position



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import io
from functools import cached_property
from io import SEEK_CUR, SEEK_END, SEEK_SET
from typing import Callable, Optional, Iterator, Union

from s3torchconnectorclient._mountpoint_s3_client import (
    ObjectInfo,
    GetObjectStream,
    HeadObjectResult,
)

from .s3reader import S3Reader

DEFAULT_BUFFER_SIZE = 8 * 1024 * 1024  # 8MB


class RangedS3Reader(S3Reader):
    """Range-based S3 reader implementation with adaptive buffering.

    Performs byte-range requests to read specific portions of S3 objects without
    downloading the entire file. Includes optional adaptive buffer to reduce S3 API
    calls for small, sequential reads while bypassing buffering for large reads.
    Optimal for sparse partial reads of large objects.

    Buffering behavior:

    * Small reads (< ``buffer_size``): Loads ``buffer_size`` bytes to buffer, copies to user
    * Large reads (>= ``buffer_size``): Direct S3 access, bypass buffer
    * Forward overlapping reads: Reuses existing buffer data if possible when read range extends beyond current buffer
    * Buffer can be disabled by setting ``buffer_size`` to 0
    * If ``buffer_size`` is None, uses default 8MB buffer

    Args:
        bucket: S3 bucket name
        key: S3 object key
        get_object_info: Callable that returns object metadata
        get_stream: Callable that returns stream for byte range requests
        buffer_size: Internal buffer size in bytes, defaults to 8MB
    """

    def __init__(
        self,
        bucket: str,
        key: str,
        get_object_info: Callable[[], Union[ObjectInfo, HeadObjectResult]],
        get_stream: Callable[[Optional[int], Optional[int]], GetObjectStream],
        buffer_size: Optional[int] = None,
    ):
        if not bucket:
            raise ValueError("Bucket should be specified")
        self._bucket = bucket
        self._key = key
        self._get_object_info = get_object_info
        self._get_stream = get_stream
        self._size: Optional[int] = None
        self._position: int = 0

        # Buffer Parameters
        self._buffer_size: int
        self._enable_buffering: bool
        if buffer_size is None:  # If None, use default buffer size
            self._buffer_size = DEFAULT_BUFFER_SIZE
            self._enable_buffering = True
        else:  # If integer, enable buffering if > 0 (0 disables buffer)
            self._buffer_size = buffer_size
            self._enable_buffering = buffer_size > 0
        # Create reusable buffer
        self._buffer: Optional[bytearray] = (
            bytearray(self._buffer_size) if self._enable_buffering else None
        )
        self._buffer_view: Optional[memoryview] = (
            memoryview(self._buffer) if self._buffer else None
        )
        # Track buffer byte range
        self._buffer_start: int = 0
        self._buffer_end: int = 0

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def key(self) -> str:
        return self._key

    @cached_property
    def _object_info(self):
        return self._get_object_info()

    def _load_buffer(self, start: int):
        """Load self._buffer with ranged request

        Args:
            start: Starting byte position to load into buffer
        """
        end = min(start + self._buffer_size, self._get_size())

        assert self._buffer_view is not None  # for type checker
        bytes_read = 0
        # Reuse buffer
        for chunk in self._get_stream(start, end):
            chunk_len = len(chunk)
            self._buffer_view[bytes_read : bytes_read + chunk_len] = chunk
            bytes_read += chunk_len

        # Update Buffer Boundaries
        self._buffer_start, self._buffer_end = start, start + bytes_read

    def _read_buffered(self, view: memoryview, start: int, end: int) -> int:
        """Read data from internal buffer into the provided memoryview.
        Loads buffer if buffer doesn't contain full data range.

        Args:
            view: Target memoryview to write data into
            start: Starting byte position in S3 object (inclusive)
            end: Ending byte position in S3 object (exclusive)

        Returns:
            int: Number of bytes read
        """
        if start < self._buffer_start or end > self._buffer_end:
            self._load_buffer(start)

        buffer_offset = start - self._buffer_start
        length = end - start

        assert self._buffer is not None  # for type checker
        view[:length] = self._buffer[buffer_offset : buffer_offset + length]

        return length

    def _read_unbuffered(self, view: memoryview, start: int, end: int) -> int:
        """Creates a range-based stream and reads bytes from S3 into a memoryview.

        Args:
            view: Target memoryview to write data into
            start: Starting byte position in S3 object (inclusive)
            end: Ending byte position in S3 object (exclusive)

        Returns:
            int: Number of bytes read
        """
        length = end - start
        bytes_read = 0
        for chunk in self._get_stream(start, end):
            # Safeguard for buffer overflow (stream size > length)
            chunk_size = min(len(chunk), length - bytes_read)
            chunk_view = memoryview(chunk)
            view[bytes_read : bytes_read + chunk_size] = chunk_view[:chunk_size]
            bytes_read += chunk_size
            # Exit if finished reading
            if bytes_read == length:
                break

        return bytes_read

    def _read_range(self, view: memoryview, start: int, end: int) -> int:
        """Reads into a memoryview from a specific byte range.
        Dispatch read request to buffered or unbuffered implementation based on request size.

        - Forward overlap optimization: When read range starts within current buffer and extends beyond it,
        first uses the overlapped portion from buffer, then processes remaining portion according to size.
        - Large (remaining) requests (< buffer_size): bypass buffering to speed up data transfer
        - Small (remaining) requests (>= buffer_size): use buffering to reduce S3 API calls, anticipating next call

        Args:
            view: Target memoryview to write data into
            start: Starting byte position in S3 object (inclusive)
            end: Ending byte position in S3 object (exclusive)

        Returns:
            int: Number of bytes read
        """

        bytes_read = 0

        # Forward overlap case: load from overlapped part in buffer first
        # Only apply when starting within buffer and extending beyond it, or exact match
        if self._buffer_start <= start < self._buffer_end <= end:
            # Read overlapped portion from buffer
            overlap_read = self._read_buffered(view, start, self._buffer_end)
            # Adjust start and view for the remaining portion
            start = self._buffer_end
            view = view[overlap_read:]
            bytes_read += overlap_read

        # Handle remaining portion / full range based on size
        if end - start >= self._buffer_size or not self._enable_buffering:
            # Large reads: return data directly
            bytes_read += self._read_unbuffered(view, start, end)
        else:
            # Small reads: buffer data and return from buffer
            bytes_read += self._read_buffered(view, start, end)

        self._position += bytes_read
        return bytes_read

    def readinto(self, buf) -> int:
        """Read up to len(buf) bytes into a pre-allocated, writable bytes-like object buf.
        Return the number of bytes read. If no bytes are available, zero is returned.

        Args:
            buf : writable bytes-like object

        Returns:
            int : numer of bytes read or zero, if no bytes available
        """

        try:
            view = memoryview(buf)
            if view.readonly:
                raise TypeError(
                    f"argument must be a writable bytes-like object, not {type(buf).__name__}"
                )
        except TypeError:
            raise TypeError(
                f"argument must be a writable bytes-like object, not {type(buf).__name__}"
            )

        buf_size = len(buf)
        if self._position_at_end() or buf_size == 0:
            # If no bytes are available or no place to write data, zero should be returned
            return 0

        # Calculate the range to request
        start = self._position
        end = min(start + buf_size, self._get_size())

        # Return no data if zero-length range
        if start >= end:
            return 0

        return self._read_range(view, start, end)

    def read(self, size: Optional[int] = None) -> bytes:
        """Read up to size bytes from the current position.

        If size is zero or positive, read that many bytes from S3, or until the end of the object.
        If size is None or negative, read until the end of the object.

        Args:
            size (int | None): how many bytes to read.

        Returns:
            bytes: Bytes read from specified range.

        Raises:
            S3Exception: An error occurred accessing S3.
        """

        if size is not None and not isinstance(size, int):
            raise TypeError(f"argument should be integer or None, not {type(size)!r}")
        if self._position_at_end():
            # Invariant: if we're at EOF, it doesn't matter what `size` is, we'll always return no data and have no
            # side effect.
            return b""

        # Calculate the range to request
        start = self._position
        if size is None or size < 0:
            end = self._get_size()
        else:
            end = min(start + size, self._get_size())

        # Return no data if zero-length range
        if start >= end:
            return b""

        # Pre-allocate buffer
        byte_size = end - start
        buffer = bytearray(byte_size)
        view = memoryview(buffer)

        self._read_range(view, start, end)
        return view.tobytes()

    def seek(self, offset: int, whence: int = SEEK_SET, /) -> int:
        """Change the stream position to the given byte offset, interpreted relative to whence.

        When seeking beyond the end of the file, always stay at EOF.
        Seeking before the start of the file results in a ValueError.

        Args:
            offset (int): How many bytes to seek relative to whence.
            whence (int): One of SEEK_SET, SEEK_CUR, and SEEK_END. Default: SEEK_SET

        Returns:
            int: Current position of the stream

        Raises:
            S3Exception: An error occurred accessing S3.

        """
        if not isinstance(offset, int):
            raise TypeError(f"integer argument expected, got {type(offset)!r}")
        if whence == SEEK_END:
            if offset >= 0:
                self._position = self._get_size()
                return self._position
            offset += self._get_size()
        elif whence == SEEK_CUR:
            if self._position_at_end() and offset >= 0:
                return self._position
            offset += self._position
        elif whence == SEEK_SET:
            pass
        elif isinstance(whence, int):
            raise ValueError("Seek must be passed SEEK_CUR, SEEK_SET, or SEEK_END")
        else:
            raise TypeError(f"integer argument expected, got {type(whence)!r}")

        if offset < 0:
            raise ValueError(f"negative seek value {offset}")

        # Update position without prefetching.
        self._position = min(offset, self._get_size())
        return self._position

    def _get_size(self) -> int:
        if self._size is None:
            self._size = self._object_info.size
        return self._size

    def _position_at_end(self) -> bool:
        # Code calling this must only be used for optimisation purposes.
        if self._size is None:
            # We can never be special cased to EOF if we never saw how long it is.
            # If we _are_ at EOF, we'll just not take the early exits.
            return False
        return self._position == self._size

    def tell(self) -> int:
        """
        Returns:
              int: Current stream position.
        """
        return self._position



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class S3ClientConfig:
    """A dataclass exposing configurable parameters for the S3 client.

    Args:
    throughput_target_gbps(float): Throughput target in Gigabits per second (Gbps) that we are trying to reach.
        10.0 Gbps by default (may change in future).
    part_size(int): Size (bytes) of file parts that will be uploaded/downloaded.
        Note: for saving checkpoints, the inner client will adjust the part size to meet the service limits.
        (max number of parts per upload is 10,000, minimum upload part size is 5 MiB).
        Part size must have values between 5MiB and 5GiB.
        8MiB by default (may change in future).
    unsigned(bool): Set to true to disable signing S3 requests.
    force_path_style(bool): forceful path style addressing for S3 client.
    max_attempts(int): amount of retry attempts for retrieable errors.
    profile(str): Profile name to use for S3 authentication.
    """

    throughput_target_gbps: float = 10.0
    part_size: int = 8 * 1024 * 1024
    unsigned: bool = False
    force_path_style: bool = False
    max_attempts: int = 10
    profile: Optional[str] = None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD
from typing import Optional

from ._s3dataset_common import parse_s3_uri
from ._s3client import S3Client, S3ClientConfig
from . import S3Reader, S3Writer


class S3Checkpoint:
    """A checkpoint manager for S3.

    To read a checkpoint from S3, users need to create an S3Reader
    by providing s3_uri of the checkpoint stored in S3. Similarly, to save a
    checkpoint to S3, users need to create an S3Writer by providing s3_uri.
    S3Reader and S3Writer implements io.BufferedIOBase therefore, they can be passed to
    torch.load, and torch.save.
    """

    def __init__(
        self,
        region: str,
        endpoint: Optional[str] = None,
        s3client_config: Optional[S3ClientConfig] = None,
    ):
        self.region = region
        self.endpoint = endpoint
        self._client = S3Client(
            region, endpoint=endpoint, s3client_config=s3client_config
        )

    def reader(self, s3_uri: str) -> S3Reader:
        """Creates an S3Reader from a given s3_uri.

        Args:
            s3_uri (str): A valid s3_uri. (i.e. s3://<BUCKET>/<KEY>)

        Returns:
            S3Reader: a read-only binary stream of the S3 object's contents, specified by the s3_uri.

        Raises:
            S3Exception: An error occurred accessing S3.
        """
        bucket, key = parse_s3_uri(s3_uri)
        return self._client.get_object(bucket, key)

    def writer(self, s3_uri: str) -> S3Writer:
        """Creates an S3Writer from a given s3_uri.

        Args:
            s3_uri (str): A valid s3_uri. (i.e. s3://<BUCKET>/<KEY>)

        Returns:
            S3Writer: a write-only binary stream. The content is saved to S3 using the specified s3_uri.

        Raises:
            S3Exception: An error occurred accessing S3.
        """
        bucket, key = parse_s3_uri(s3_uri)
        return self._client.put_object(bucket, key)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import logging
import os
from functools import partial
from typing import TYPE_CHECKING, Optional, List, Dict, Union
from collections import defaultdict

from .s3reader import S3Reader
from .protocol import (
    S3ReaderConstructorProtocol,
    DCPS3ReaderConstructorProtocol,
)
from .sequential import SequentialS3Reader
from .ranged import RangedS3Reader
from .dcp_optimized import (
    DCPOptimizedS3Reader,
    ItemRange,
    DEFAULT_MAX_GAP_SIZE,
    FALLBACK_GUIDANCE,
)

if TYPE_CHECKING:
    from torch.distributed.checkpoint.planner import ReadItem
    from torch.distributed.checkpoint.metadata import MetadataIndex
    from torch.distributed.checkpoint.filesystem import _StorageInfo

log = logging.getLogger(__name__)


class DCPOptimizedConstructor:
    """Constructor for DCPOptimizedS3Reader instances with range metadata injection.

    Created from S3ReaderConstructor, and used by S3StorageReader for PyTorch DCP.
    Requires range metadata from DCP load plans to function properly.

    Usage Flow:
        # User Setup
        reader_constructor = S3ReaderConstructor.dcp_optimized()  -> DCPOptimizedConstructor
        S3StorageReader(..., reader_constructor)                  -> stores in S3FileSystem

        # During DCP.load()
        S3StorageReader.prepare_local_plan(plan)
            -> set_item_ranges_by_file(plan.items, storage_data, path)
            -> builds _item_ranges_by_file: {s3_uri: [ItemRange, ...]}

        # During read_data(), per file
        S3FileSystem.create_stream(path, "rb")
            -> S3Client.get_object(bucket, key, reader_constructor)
                -> __call__(bucket, key, get_object_info, get_stream)
                    -> .metadata: SequentialS3Reader (no ranges available for .metadata file)
                    -> .distcp:   DCPOptimizedS3Reader(item_ranges) ranges from _item_ranges_by_file
    """

    def __init__(self, max_gap_size: Union[int, float] = DEFAULT_MAX_GAP_SIZE) -> None:

        if max_gap_size < 0:
            raise ValueError("max_gap_size must be non-negative")

        self._item_ranges_by_file: Dict[str, List[ItemRange]] = {}
        self._max_gap_size: Union[int, float] = max_gap_size

    def set_item_ranges_by_file(
        self,
        plan_items: "List[ReadItem]",
        storage_data: "Dict[MetadataIndex, _StorageInfo]",
        base_path: Union[str, os.PathLike],
    ) -> None:
        """Extract and store item ranges from DCP load plan.
        Called by S3StorageReader.prepare_local_plan() to inject range metadata.

        Note: This replaces any previously stored ranges (intentional for multi-call scenarios).
        """

        if not plan_items:
            # Empty plan: no reads will happen since no-op in FileSystemReader.read_data
            # To defend against multi-processing case where a rank doesn't need data.
            log.debug("Empty plan for this rank - no DCP item ranges will be created.")
            self._item_ranges_by_file = defaultdict(list)
            return

        if not storage_data:
            raise ValueError(
                "storage_data must not be empty; required to map ReadItems to file ranges."
            )

        # Map: full S3 URI -> item ranges
        self._item_ranges_by_file = defaultdict(list)
        for read_item in plan_items:
            item_md = storage_data[read_item.storage_index]
            s3_uri = os.path.join(base_path, item_md.relative_path)
            self._item_ranges_by_file[s3_uri].append(
                ItemRange(item_md.offset, item_md.offset + item_md.length)
            )

    def __call__(self, bucket: str, key: str, get_object_info, get_stream) -> S3Reader:
        """Match key to corresponding List[ItemRange] in _item_ranges_by_file.
        Fallback when reading .metadata file if ranges are unavailable.
        """

        # Return SequentialS3Reader for .metadata (hardcoded name in FileSystemReader.read_metadata),
        # since there are no ranges available for .metadata.
        # TODO: alternatively configure DCPOptimizedS3Reader to use full file range for .metadata.
        if key.endswith(".metadata"):
            log.debug(f"Reading .metadata file {key} with SequentialS3Reader")
            return SequentialS3Reader(bucket, key, get_object_info, get_stream)

        # Return DCPOptimizedS3Reader for .distcp files with corresponding item ranges for file uri
        s3_uri = f"s3://{bucket}/{key}"
        if s3_uri in self._item_ranges_by_file:
            return DCPOptimizedS3Reader(
                bucket,
                key,
                item_ranges=self._item_ranges_by_file[s3_uri],
                get_object_info=get_object_info,
                get_stream=get_stream,
                max_gap_size=self._max_gap_size,
            )

        # Error for other files; warn users in case they override prepare_local_plan behavior
        raise ValueError(
            f"No ranges found for {s3_uri}. Make sure range injection is used in "
            f"'S3StorageReader.prepare_local_plan'.\n{FALLBACK_GUIDANCE}"
        )


class S3ReaderConstructor:
    """Constructor for creating ``partial(S3Reader)`` instances.

    Creates partial ``S3Reader`` instances that will be completed by ``S3Client`` with the
    remaining required parameters (e.g. ``bucket``, ``key``, ``get_object_info``, ``get_stream``).

    The constructor provides factory methods for different reader types:

    - ``sequential()``: Creates a constructor for sequential readers that buffer the entire object.
      Best for full reads and repeated access.
    - ``range_based()``: Creates a constructor for range-based readers that fetch specific byte ranges.
      Suitable for sparse partial reads for large objects.
    """

    @staticmethod
    def sequential() -> S3ReaderConstructorProtocol:
        """Creates a constructor for sequential (generic) readers.

        This reader is the generic reader that supports all access patterns.

        Returns:
            S3ReaderConstructorProtocol: Partial constructor for SequentialS3Reader

        Example::

            reader_constructor = S3ReaderConstructor.sequential()

        """
        return partial(SequentialS3Reader)

    @staticmethod
    def range_based(buffer_size: Optional[int] = None) -> S3ReaderConstructorProtocol:
        """Creates a constructor for range-based readers

        Args:
            buffer_size: Internal buffer size in bytes. If None, uses default 8MB.
                         Set to 0 to disable buffering.

        Returns:
            S3ReaderConstructorProtocol: Partial constructor for RangedS3Reader

        Range-based reader performs byte-range requests for each read/readinto call
        to read specific portions of S3 objects without downloading the entire file.

        Buffer size affects read performance:

        * Small reads (< ``buffer_size``): Loads ``buffer_size`` bytes to buffer to reduce S3 API calls for small, sequential reads
        * Large reads (≥ ``buffer_size``): bypass the buffer for direct transfer from S3
        * Forward overlap reads: Reuses buffered data when reading ranges that extend beyond current buffer, and processes remaining
        data according to size with logic above.

        Configuration Guide:

        * Use larger buffer sizes for workloads with many small, sequential reads of nearby bytes
        * Use smaller buffer sizes or disable buffering for sparse partial reads
        * Buffer can be disabled by setting ``buffer_size`` to 0
        * If ``buffer_size`` is None, uses default 8MB buffer

        Examples::

            # Range-based reader with default 8MB buffer
            reader_constructor = S3ReaderConstructor.range_based()

            # Range-based reader with custom buffer size
            reader_constructor = S3ReaderConstructor.range_based(buffer_size=16*1024*1024)

            # Range-based reader with buffering disabled
            reader_constructor = S3ReaderConstructor.range_based(buffer_size=0)
        """
        return partial(RangedS3Reader, buffer_size=buffer_size)

    @staticmethod
    def dcp_optimized(
        max_gap_size: Union[int, float] = DEFAULT_MAX_GAP_SIZE,
    ) -> DCPS3ReaderConstructorProtocol:
        """Creates a constructor for DCP-optimized readers for faster checkpoint loading.

        The DCP-optimized reader provides performance improvements for DCP reading through:

        - Selective data fetching with range coalescing to only fetch required byte ranges
        - Per-item buffer management to reduce buffer allocation costs
        - Eliminating buffer copy by storing S3 chunks as memoryview references

        Args:
            max_gap_size: Maximum gap size in bytes between ranges to coalesce into the same S3 read stream.
                Most users should use the default value.

                - Default: 32MB (``32 * 1024 * 1024``)
                - Use ``float("inf")`` to coalesce all ranges regardless of gaps
                - Use 0 to disable coalescing, which creates a new range-based stream for each gap

        Returns:
            DCPOptimizedConstructorProtocol:
                Constructor that creates DCPOptimizedS3Reader when ranges are available, falling back to
                SequentialS3Reader otherwise.

        Requirements:
            Should be used with S3StorageReader, in which ``prepare_local_plan()`` automatically handles:

            - Load ordering: Sorts items by storage offset for sequential access
            - Range injection: Provides byte ranges from DCP load plan to the reader

            Advanced users implementing custom readers must include these optimizations
            in their ``prepare_local_plan()``/``read_data()`` implementation to use the DCP-optimized reader.

        Example::

            reader_constructor = S3ReaderConstructor.dcp_optimized()
            storage_reader = S3StorageReader(region, path, reader_constructor=reader_constructor)
            DCP.load(state_dict, storage_reader=storage_reader)

        """
        return DCPOptimizedConstructor(max_gap_size=max_gap_size)

    @staticmethod
    def default() -> S3ReaderConstructorProtocol:
        """Creates the default generic reader constructor.

        This creates a sequential (generic) reader that supports all access patterns.

        Returns:
            S3ReaderConstructorProtocol: Partial constructor for SequentialS3Reader
        """
        return S3ReaderConstructor.sequential()

    @staticmethod
    def get_reader_type_string(
        constructor: Optional[S3ReaderConstructorProtocol],
    ) -> str:
        """Returns the reader type string for the given constructor."""
        if constructor is None:
            return S3ReaderConstructor.get_reader_type_string(
                S3ReaderConstructor.default()
            )

        if isinstance(constructor, DCPOptimizedConstructor):
            return "dcp_optimized"
        elif not isinstance(constructor, partial):
            return "unknown"
        elif constructor.func == RangedS3Reader:
            return "range_based"
        elif constructor.func == SequentialS3Reader:
            return "sequential"
        else:
            return "unknown"



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

from .s3reader import S3Reader
from .constructor import S3ReaderConstructor, DCPOptimizedConstructor
from .sequential import SequentialS3Reader
from .ranged import RangedS3Reader
from .dcp_optimized import DCPOptimizedS3Reader, ItemRange, RangeGroup
from .protocol import (
    GetStreamCallable,
    S3ReaderConstructorProtocol,
    DCPS3ReaderConstructorProtocol,
)

__all__ = [
    "S3Reader",
    "S3ReaderConstructor",
    "SequentialS3Reader",
    "RangedS3Reader",
    "DCPOptimizedS3Reader",
]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD
import platform
from typing import List, Optional

import torch

from ._version import __version__

# https://www.rfc-editor.org/rfc/rfc9110#name-user-agent


class UserAgent:
    @staticmethod
    def get_default_prefix() -> str:
        """
        Get the default user agent prefix without any comments.

        Format: s3torchconnector/{version} ua/2.1 os/{os}#{version}
                lang/python#{version} md/arch#{arch} md/pytorch#{version}

        Returns:
            str: The default user agent prefix string
        """
        os_name = platform.system().lower()
        if os_name == "darwin":
            os_name = "macos"
        os_version = platform.release()
        python_version = platform.python_version()
        arch = platform.machine().lower()
        pytorch_version = torch.__version__

        parts = [
            f"{__package__}/{__version__}",
            "ua/2.1",
            f"os/{os_name}#{os_version}",
            f"lang/python#{python_version}",
            f"md/arch#{arch}",
            f"md/pytorch#{pytorch_version}",
        ]

        return " ".join(parts)

    def __init__(self, comments: Optional[List[str]] = None):
        if comments is not None and not isinstance(comments, list):
            raise ValueError("Argument comments must be a List[str]")

        self._user_agent_prefix = UserAgent.get_default_prefix()
        self._comments = comments or []

    @property
    def prefix(self):
        comments_str = "; ".join(filter(None, self._comments))
        if comments_str:
            return f"{self._user_agent_prefix} ({comments_str})"
        return self._user_agent_prefix



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

from typing import Iterable, Union, Tuple

from ._s3_bucket_iterable import S3BucketIterable
from ._s3client import S3Client
from . import S3Reader
from ._s3bucket_key_data import S3BucketKeyData

"""
_s3dataset_common.py
    Collection of common methods for S3 datasets, containing logic for URIs parsing and objects listing. 
"""


def identity(obj: S3Reader) -> S3Reader:
    return obj


# TODO: Check boto3 implementation for this
def parse_s3_uri(uri: str) -> Tuple[str, str]:
    if not uri or not uri.startswith("s3://"):
        raise ValueError("Only s3:// URIs are supported")
    uri = uri[len("s3://") :]
    if not uri:
        raise ValueError("Bucket name must be non-empty")
    split = uri.split("/", maxsplit=1)
    if len(split) == 1:
        bucket = split[0]
        prefix = ""
    else:
        bucket, prefix = split
    if not bucket:
        raise ValueError("Bucket name must be non-empty")
    return bucket, prefix


def get_objects_from_uris(
    object_uris: Union[str, Iterable[str]], client: S3Client
) -> Iterable[S3BucketKeyData]:
    if isinstance(object_uris, str):
        object_uris = [object_uris]
    # TODO: We should be consistent with URIs parsing. Revise if we want to do this upfront or lazily.
    bucket_key_pairs = [parse_s3_uri(uri) for uri in object_uris]

    return (S3BucketKeyData(bucket, key) for bucket, key in bucket_key_pairs)


def get_objects_from_prefix(s3_uri: str, client: S3Client) -> Iterable[S3BucketKeyData]:
    bucket, prefix = parse_s3_uri(s3_uri)
    return iter(S3BucketIterable(client, bucket, prefix))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import io
from io import SEEK_SET
from abc import ABC, abstractmethod
from typing import Optional


class S3Reader(ABC, io.BufferedIOBase):
    """An abstract base class for read-only, file-like representation of a single object stored in S3.

    This class defines the interface for S3 readers. Concrete implementations (SequentialS3Reader or
    RangedS3Reader extend this class. S3ReaderConstructor creates partial functions of these
    implementations, which are then completed by S3Client with the remaining required parameters.
    """

    @property
    @abstractmethod
    def bucket(self) -> str:
        pass

    @property
    @abstractmethod
    def key(self) -> str:
        pass

    @abstractmethod
    def read(self, size: Optional[int] = None) -> bytes:
        pass

    @abstractmethod
    def seek(self, offset: int, whence: int = SEEK_SET, /) -> int:
        pass

    @abstractmethod
    def tell(self) -> int:
        pass

    @abstractmethod
    def readinto(self, buf) -> int:
        pass

    def seekable(self) -> bool:
        """
        Returns:
            bool: Return whether object supports seek operations.
        """
        return True

    def readable(self) -> bool:
        """
        Returns:
            bool: Return whether object was opened for reading.
        """
        return True

    def writable(self) -> bool:
        """
        Returns:
            bool: Return whether object was opened for writing.
        """
        return False



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD
from functools import partial
from typing import Iterator, Any, Union, Iterable, Callable, Optional
import logging

import torch.utils.data
import torch

from . import S3Reader, S3ReaderConstructor
from .s3reader import S3ReaderConstructorProtocol
from ._s3bucket_key_data import S3BucketKeyData
from ._s3client import S3Client, S3ClientConfig
from ._user_agent import UserAgent
from ._s3dataset_common import (
    identity,
    get_objects_from_uris,
    get_objects_from_prefix,
)

log = logging.getLogger(__name__)


class S3IterableDataset(torch.utils.data.IterableDataset):
    """An IterableStyle dataset created from S3 objects.

    To create an instance of S3IterableDataset, you need to use
    `from_prefix` or `from_objects` methods.
    """

    def __init__(
        self,
        region: str,
        get_dataset_objects: Callable[[S3Client], Iterable[S3BucketKeyData]],
        endpoint: Optional[str] = None,
        transform: Callable[[S3Reader], Any] = identity,
        s3client_config: Optional[S3ClientConfig] = None,
        enable_sharding: bool = False,
        reader_constructor: Optional[S3ReaderConstructorProtocol] = None,
    ):
        self._get_dataset_objects = get_dataset_objects
        self._transform = transform
        self._region = region
        self._endpoint = endpoint
        self._s3client_config = s3client_config
        self._client = None
        self._enable_sharding = enable_sharding
        self._reader_constructor = reader_constructor or S3ReaderConstructor.default()

        self._rank = 0
        self._world_size = 1
        if torch.distributed.is_initialized():
            self._rank = torch.distributed.get_rank()
            self._world_size = torch.distributed.get_world_size()

    @property
    def region(self):
        return self._region

    @property
    def endpoint(self):
        return self._endpoint

    @classmethod
    def from_objects(
        cls,
        object_uris: Union[str, Iterable[str]],
        *,
        region: str,
        endpoint: Optional[str] = None,
        transform: Callable[[S3Reader], Any] = identity,
        s3client_config: Optional[S3ClientConfig] = None,
        enable_sharding: bool = False,
        reader_constructor: Optional[S3ReaderConstructorProtocol] = None,
    ):
        """Returns an instance of S3IterableDataset using the S3 URI(s) provided.

        Args:
          object_uris(str | Iterable[str]): S3 URI of the object(s) desired.
          region(str): AWS region of the S3 bucket where the objects are stored.
          endpoint(str): AWS endpoint of the S3 bucket where the objects are stored.
          transform: Optional callable which is used to transform an S3Reader into the desired type.
          s3client_config: Optional S3ClientConfig with parameters for S3 client.
          enable_sharding: If True, shard the dataset across multiple workers for parallel data loading. If False (default), each worker loads the entire dataset independently.
          reader_constructor (Optional[S3ReaderConstructorProtocol]): Optional partial(S3Reader) created using S3ReaderConstructor
            e.g. S3ReaderConstructor.sequential() or S3ReaderConstructor.range_based()

        Returns:
            S3IterableDataset: An IterableStyle dataset created from S3 objects.

        Raises:
            S3Exception: An error occurred accessing S3.
        """
        log.info(f"Building {cls.__name__} from_objects")
        return cls(
            region,
            partial(get_objects_from_uris, object_uris),
            endpoint,
            transform=transform,
            s3client_config=s3client_config,
            enable_sharding=enable_sharding,
            reader_constructor=reader_constructor,
        )

    @classmethod
    def from_prefix(
        cls,
        s3_uri: str,
        *,
        region: str,
        endpoint: Optional[str] = None,
        transform: Callable[[S3Reader], Any] = identity,
        s3client_config: Optional[S3ClientConfig] = None,
        enable_sharding: bool = False,
        reader_constructor: Optional[S3ReaderConstructorProtocol] = None,
    ):
        """Returns an instance of S3IterableDataset using the S3 URI provided.

        Args:
          s3_uri(str): An S3 URI (prefix) of the object(s) desired. Objects matching the prefix will be included in the returned dataset.
          region(str): AWS region of the S3 bucket where the objects are stored.
          endpoint(str): AWS endpoint of the S3 bucket where the objects are stored.
          transform: Optional callable which is used to transform an S3Reader into the desired type.
          s3client_config: Optional S3ClientConfig with parameters for S3 client.
          enable_sharding: If True, shard the dataset across multiple workers for parallel data loading. If False (default), each worker loads the entire dataset independently.
          reader_constructor (Optional[S3ReaderConstructorProtocol]): Optional partial(S3Reader) created using S3ReaderConstructor
            e.g. S3ReaderConstructor.sequential() or S3ReaderConstructor.range_based()

        Returns:
            S3IterableDataset: An IterableStyle dataset created from S3 objects.

        Raises:
            S3Exception: An error occurred accessing S3.
        """
        log.info(f"Building {cls.__name__} from_prefix {s3_uri=}")
        return cls(
            region,
            partial(get_objects_from_prefix, s3_uri),
            endpoint,
            transform=transform,
            s3client_config=s3client_config,
            enable_sharding=enable_sharding,
            reader_constructor=reader_constructor,
        )

    def _get_client(self):
        if self._client is None:
            reader_type_string = S3ReaderConstructor.get_reader_type_string(
                self._reader_constructor
            )
            self._client = S3Client(
                self.region,
                endpoint=self.endpoint,
                user_agent=UserAgent(
                    comments=[
                        f"md/dataset#iterable md/reader_type#{reader_type_string}"
                    ]
                ),
                s3client_config=self._s3client_config,
            )
        return self._client

    def _get_transformed_object(self, bucket_key: S3BucketKeyData) -> Any:
        return self._transform(
            self._get_client().get_object(
                bucket_key.bucket,
                bucket_key.key,
                object_info=bucket_key.object_info,
                reader_constructor=self._reader_constructor,
            )
        )

    def __iter__(self) -> Iterator[Any]:
        worker_id = 0
        num_workers = 1
        if self._enable_sharding:
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                worker_id = worker_info.id
                num_workers = worker_info.num_workers

        if not self._enable_sharding or (self._world_size == 1 and num_workers == 1):
            # sharding disabled or only one shard is available, so return the entire dataset
            return map(
                self._get_transformed_object,
                self._get_dataset_objects(self._get_client()),
            )

        """In a multi-process setting (e.g., distributed training), the dataset needs to be
        sharded across multiple processes. The following variables control this sharding:

        _rank: The rank (index) of the current process within the world (group of processes).
        _world_size: The total number of processes in the world (group).

        In addition, within each process, the dataset may be further sharded across multiple
        worker threads or processes (e.g., for data loading). The following variables control
        this intra-process sharding:

        worker_id: The ID of the current worker thread/process within the process.
        num_workers: The total number of worker threads/processes within the process.
        """

        # First, distribute objects across ranks
        rank_sharded_objects = (
            obj
            for idx, obj in enumerate(self._get_dataset_objects(self._get_client()))
            if idx % self._world_size == self._rank
        )

        # Then, distribute objects within each rank across workers
        worker_sharded_objects = (
            obj
            for idx, obj in enumerate(rank_sharded_objects)
            if idx % num_workers == worker_id
        )

        return map(self._get_transformed_object, worker_sharded_objects)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

"""
DCP-Optimized S3 Reader provides these 3 optimizations:
1. Selective data fetching with range coalescing to only fetch required byte ranges
2. Per-item buffer management to reduce buffer allocation costs
3. Eliminating buffer copy by storing S3 chunks as memoryview references

Data Flow Overview:
    DCP.load(model_state_dict, storage_reader=s3_storage_reader)
        -> read_metadata()                                          # reads .metadata file
        -> set_up_storage_reader(metadata)                          # populates storage_data
        -> prepare_local_plan(plan)                                 # (patched) sorts items, injects ranges to constructor
            -> DCPOptimizedConstructor.set_item_ranges_by_file()
        -> read_data(plan)                                          # per-file loop below
            -> DCPOptimizedS3Reader __init__
                -> _validate_and_coalesce_ranges()                  # validates and groups ItemRanges into RangeGroups
            -> DCPOptimizedS3Reader read()/readinto()
                -> _find_item_for_range()                           # returns item for given read request
                -> [if new item] _get_item_buffer()                 # fetches item byte data
                    -> [if new RangeGroup] _get_stream_for_item()   # creates new stream before fetching byte data
                    -> 1: Handle leftover bytes from prev. chunk
                    -> 2: Skip gap data from coalescing
                    -> 3: Fetch remaining data from S3
                -> _ItemViewBuffer read()/readinto()                # returns data from buffer
"""

import bisect
import logging
from dataclasses import dataclass
from typing import List, Optional, Callable, Union, Iterator, Dict, cast
from io import SEEK_SET, SEEK_CUR, SEEK_END

from s3torchconnectorclient._mountpoint_s3_client import (
    ObjectInfo,
    GetObjectStream,
    HeadObjectResult,
)
from .s3reader import S3Reader

log = logging.getLogger(__name__)

DEFAULT_MAX_GAP_SIZE = (
    32 * 1024 * 1024
)  # default max gap between ranges such that they stay in the same stream
FIND_ITEM_ERROR_PREFIX = (
    "DCPOptimizedS3Reader only supports sequentially accessing provided ranges: "
)
FALLBACK_GUIDANCE = (
    "If this error is encountered with the default DCP reader (S3ReaderConstructor.dcp_optimized()) "
    "added in s3torchconnector v1.5.0, please refer to the troubleshooting doc "
    "(https://github.com/awslabs/s3-connector-for-pytorch/blob/main/docs/TROUBLESHOOTING.md#dcpoptimizeds3reader-errors)."
    "\nFor unsupported or non-DCP access patterns, use the generic reader: "
    "S3StorageReader(region, path, reader_constructor=S3ReaderConstructor.default())"
)


@dataclass
class ItemRange:
    """Byte range for a single DCP ReadItem (tensor). Inclusive start, exclusive end."""

    start: int
    end: int


@dataclass
class RangeGroup:
    """Group of nearby ItemRanges that will share a single S3 range request.

    Created by coalescing ItemRanges with gaps <= max_gap_size in _validate_and_coalesce_ranges.
    One S3 stream will serve all items in the RangeGroup sequentially.
    """

    start: int  # First byte of the group (= first item's start)
    end: int  # Last byte of the group (= last item's end)
    item_ranges: List[ItemRange]  # Items within this group, in order


@dataclass
class _StreamState:
    """
    Tracks S3 stream position in current RangeGroup and buffered data between item reads.

    A single S3 stream may serve multiple items in a RangeGroup. This state tracks
        1. Where we are in that stream created for the current RangeGroup, and
        2. any leftover bytes from the previous chunk that belong to the next item.
    """

    stream: Optional[GetObjectStream] = None  # Current S3 stream for active RangeGroup
    stream_position: int = (
        -1
    )  # Current byte position in S3 stream; -1 as dummy init value
    leftover: Optional[memoryview] = None  # Unused bytes from end of last chunk read


# TODO: extend buffer for use in other S3Reader implementations after extensive testing
class _ItemViewBuffer:
    """
    A read-only buffer storing item data, in the form of multiple memoryview segments.

    Instead of copying S3 chunks into a growing BytesIO buffer, this class stores
    references to the original S3 chunks (typically 8MB parts) as memoryview segments.
    This allows us to reduce buffer allocation costs, saving time and memory.

    The buffer supports seek/read/readinto with logic that handles reads spanning
    multiple segments, similar to the file-access interface in io.BytesIO.
    """

    __slots__ = ("_segments", "_offsets", "_lengths", "_size", "_pos")

    def __init__(self) -> None:
        self._segments: List[memoryview] = []  # memoryview segments
        self._offsets: List[int] = []  # start offset (within the item) of each segment
        self._lengths: List[int] = (
            []
        )  # length of each segment (avoid recalculations from offset)
        self._size: int = 0  # total item length (sum of _lengths)
        self._pos: int = 0  # current read position within the item

    def append_view(self, view: memoryview) -> None:
        """Append a memoryview segment (ignored if empty)."""

        seg_len = len(view)
        if seg_len == 0:
            return
        self._segments.append(view)
        self._offsets.append(self._size)
        self._lengths.append(seg_len)
        self._size += seg_len

    def seek(self, offset: int, whence: int = SEEK_SET, /) -> int:
        assert isinstance(offset, int), f"integer expected, got {type(offset)!r}"

        if whence == SEEK_SET:
            new_pos = offset
        elif whence == SEEK_CUR:
            new_pos = self._pos + offset
        elif whence == SEEK_END:
            new_pos = self._size + offset
        else:
            raise ValueError("Seek must be passed SEEK_CUR, SEEK_SET, or SEEK_END")

        assert new_pos >= 0, f"negative seek value {new_pos}"

        # Seeking past EOF is allowed.
        self._pos = new_pos
        return self._pos

    def tell(self) -> int:
        """Return the current pos position (like BytesIO.tell)."""
        return self._pos

    def read(self, size: Optional[int] = None) -> bytes:
        """Returns byte copy of data from the buffer, using readinto() logic.

        Note that in DCP, only PyTorch's serialization.py:: _is_zipfile() magic
        number check (read(4)) uses read() instead of readinto().
        """

        # DCPOptimizedS3Reader doesn't allow full read, and doesn't use full reads on items either.
        assert size is not None, "Size cannot be None; full read is not supported"
        assert size >= 0, "Size cannot be negative; full read is not supported"

        # Fast path: PyTorch's serialization.py::_is_zipfile() reads first 4 bytes to check magic number.
        if size == 4 and self._pos == 0 and self._lengths and self._lengths[0] >= 4:
            self._pos = 4
            # TODO: eliminating bytes() conversion can save ~3% time? Requires interface changes.
            return bytes(self._segments[0][:4])

        if size == 0:
            return b""

        # Pass implementation to readinto()
        out = bytearray(size)
        n = self.readinto(out)
        return bytes(out) if n == size else memoryview(out)[:n].tobytes()

    def readinto(self, buf) -> int:
        """Read into pre-allocated buffer, copying across segment boundaries as needed."""
        # Avoid creating new memoryview if input already is one
        dest = buf if isinstance(buf, memoryview) else memoryview(buf)
        assert not dest.readonly, "writable buffer required"

        dest_len = len(dest)
        size = self._size
        pos = self._pos

        if dest_len == 0 or pos >= size:
            return 0

        # Cache to avoid repeated attribute lookups in the loop
        segments = self._segments
        offsets = self._offsets
        lengths = self._lengths

        # Find segment containing pos with binary search
        # bisect_right gives insertion point where pos < offsets[i], -1 gives containing segment.
        # No caching optimisation, since torch.load jumps around (magic bytes, zip dir, tensor data)
        seg_idx = bisect.bisect_right(offsets, pos) - 1
        assert seg_idx >= 0, f"Invalid segment index {seg_idx} for position {pos}"

        written = 0
        bytes_to_read = min(dest_len, size - pos)

        # Copy from segments to dest, handling segment boundaries
        while written < bytes_to_read:
            seg_start = offsets[seg_idx]
            seg_len = lengths[seg_idx]
            seg = segments[seg_idx]

            # Account for first chunk when pos > seg_start
            offset_in_seg = pos - seg_start

            # Account for last chunk when bytes_to_read < seg_len
            available_in_seg = seg_len - offset_in_seg
            bytes_left_to_read = bytes_to_read - written
            copy_size = min(bytes_left_to_read, available_in_seg)

            dest[written : written + copy_size] = seg[
                offset_in_seg : offset_in_seg + copy_size
            ]

            written += copy_size
            pos += copy_size
            seg_idx += 1

        self._pos += written
        return written


class DCPOptimizedS3Reader(S3Reader):
    """S3 reader implementation optimized for PyTorch Distributed Checkpoint (DCP) loading.

    Provides up to 2x performance improvement over default sequential reader through:

        1. **Selective data fetching with range coalescing**: Uses byte range information from
        PyTorch's ``LoadPlan`` to only fetch required data. Groups nearby ranges within
        ``max_gap_size`` into single S3 streams to minimize first-byte latency while avoiding
        unnecessary data transfer.

        2. **Per-item buffer management**: Buffers per-item (per-tensor) instead of per-file. Each
        buffer stores only the required item's byte ranges and is discarded after PyTorch reads the
        item, which removes overhead of resizing large buffers and re-copying data repeatedly.

        3. **Eliminate buffer copy**: Custom ``_ItemViewBuffer`` stores S3 chunks as memoryview
        references instead of copying into BytesIO, avoiding allocation and copy overhead.

    **Requirements**:

    - DCP Loading - reader is only designed for usage via dcp_optimized reader_constructor for ``dcp.load()``
    - Pre-sorted list of item_ranges, injected automatically in ``prepare_local_plan``.
    - Sequential Access over exact item_ranges provided, also applied automatically by ``prepare_local_plan``

    **Usage**:
    Created automatically by ``DCPOptimizedConstructor`` when used with ``S3StorageReader`` and
    ``S3ReaderConstructor.dcp_optimized()``:

        reader_constructor = S3ReaderConstructor.dcp_optimized(max_gap_size=32*1024*1024)
        storage_reader = S3StorageReader(region, path, reader_constructor=reader_constructor)
        DCP.load(state_dict, storage_reader=storage_reader)

    **Error Handling**:
        Non-sequential access attempts raise ValueError.
    """

    def __init__(
        self,
        bucket: str,
        key: str,
        item_ranges: List[ItemRange],
        get_object_info: Callable[[], Union[ObjectInfo, HeadObjectResult]],
        get_stream: Callable[[Optional[int], Optional[int]], GetObjectStream],
        max_gap_size: Union[int, float] = DEFAULT_MAX_GAP_SIZE,
        # added float type to allow float("inf") / sys.maxsize for max_gap_size
    ):
        if not bucket:
            raise ValueError("Bucket should be specified")
        if not key:
            raise ValueError("Key should be specified")
        if not item_ranges:
            raise ValueError("item_ranges must be a non-empty List[ItemRange] object")
        if not isinstance(max_gap_size, (int, float)):
            raise TypeError(
                f"max_gap_size must be int or float, got {type(max_gap_size).__name__}"
            )
        if max_gap_size < 0:
            raise ValueError("max_gap_size must be non-negative")

        self._bucket = bucket
        self._key = key
        self._get_object_info = get_object_info
        self._get_stream = get_stream
        self._max_gap_size = max_gap_size
        self._closed = False

        # --- Range Processing ---

        # Filter zero-length ranges
        self._item_ranges: List[ItemRange] = [
            r for r in item_ranges if r.end != r.start
        ]
        if not self._item_ranges:
            raise ValueError("No non-empty ranges to read (all ranges were length 0)")

        # Coalesce nearby ranges into range groups that share S3 streams
        # _group_start_to_group: lookup dict for O(1) "is this item first in its group?" check
        self._group_start_to_group: Dict[int, RangeGroup] = {}
        self._range_groups: List[RangeGroup] = self._validate_and_coalesce_ranges(
            self._item_ranges, self._max_gap_size
        )

        # --- States ---

        # Stream state (stores S3 stream, position, and leftover bytes between item reads)
        self._stream_state: _StreamState = _StreamState()

        # Item buffer state
        self._item_iter: Iterator[ItemRange] = iter(
            self._item_ranges
        )  # sequential access
        self._current_item: ItemRange = next(self._item_iter)
        self._current_item_buffer: Optional[_ItemViewBuffer] = None

        # Current position in the overall S3 object
        self._position: int = 0

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def key(self) -> str:
        return self._key

    @property
    def closed(self) -> bool:
        """
        Returns:
            bool: Return whether the object is closed.
        """
        return self._closed

    def _validate_and_coalesce_ranges(
        self,
        ranges: List[ItemRange],
        max_gap_size: Union[int, float],
    ) -> List[RangeGroup]:
        """Validate ranges and coalesce nearby ranges into RangeGroups.

        - Validate: 1/ start<=end, 2/ non-negative, 3/ sorted by start position, 4/ non-overlapping
        - Coalesce: Group nearby ranges where gap <= max_gap_size into RangeGroup (one S3 stream).
        """
        if not ranges:
            return []

        groups: List[RangeGroup] = []
        items: List[ItemRange] = [ranges[0]]

        if not 0 <= ranges[0].start <= ranges[0].end:
            raise ValueError(f"Invalid range: {ranges[0].start}-{ranges[0].end}")
        for r in ranges[1:]:
            if r.end <= r.start:  # Empty ranges filtered out in __init__
                raise ValueError(f"Invalid range: {r.start}-{r.end}")
            if r.start < items[-1].end:
                if r.start < items[-1].start:
                    raise ValueError(
                        f"Unsorted ranges: {items[-1].start}-{items[-1].end} and {r.start}-{r.end}"
                    )
                else:
                    raise ValueError(
                        f"Overlapping ranges: {items[-1].start}-{items[-1].end} and {r.start}-{r.end}"
                    )
            # Coalesce or create new group
            if r.start - items[-1].end <= max_gap_size:
                items.append(r)
            else:
                group = RangeGroup(items[0].start, items[-1].end, items)
                groups.append(group)
                self._group_start_to_group[items[0].start] = group
                items = [r]

        final_group = RangeGroup(items[0].start, items[-1].end, items)
        self._group_start_to_group[items[0].start] = final_group
        groups.append(final_group)
        return groups

    def _find_item_for_range(self, start: int, end: int) -> ItemRange:
        """Find which item contains the requested range [start,end), enforcing sequential access.

        Returns current item if range is within it, and advances to / returns next item if the
        range is within the next item.

        Raises human-readable errors if the range is partially/fully outside of current or next items.
        """

        item = self._current_item

        # Check if requested range is within current item
        if item.start <= start and end <= item.end:
            return item

        # 1. If start < item.end and range not within current item, raise error
        # 2. Protection against reading 2nd item before first item by checking if buffer
        # contains any data (since we initialize with 1st item instead of None)
        if start < item.end or self._current_item_buffer is None:
            raise ValueError(
                f"{FIND_ITEM_ERROR_PREFIX}Range {start}-{end} not contained in "
                f"current item {item.start}-{item.end}.\n{FALLBACK_GUIDANCE}"
            )

        # Advance to next item
        prev_item = item
        try:
            item = next(self._item_iter)
        except StopIteration:
            raise ValueError(
                f"{FIND_ITEM_ERROR_PREFIX}Range {start}-{end} not contained in last item "
                f"with range {prev_item.start}-{prev_item.end}.\n{FALLBACK_GUIDANCE}"
            )

        # Check if requested range is within next item
        if item.start <= start and end <= item.end:
            return item

        raise ValueError(
            f"{FIND_ITEM_ERROR_PREFIX}Range {start}-{end} not contained in "
            f"current item {prev_item.start}-{prev_item.end} nor the "
            f"next item {item.start}-{item.end}.\n{FALLBACK_GUIDANCE}"
        )

    def _get_stream_for_item(self, item: ItemRange) -> GetObjectStream:
        """Get or create S3 stream for the given item.

        Creates new stream if item is first in its RangeGroup, otherwise reuse stream.

        Each RangeGroup maps to a contiguous byte range in S3, and items within a
        RangeGroup are read sequentially from the same stream.

        Sequential access is already enforced in _find_item_for_range, which runs
        before _get_item_buffer (which calls this method). Reading the first item will
        trigger stream creation, and subsequent reads will simply reuse the stream.
        """

        # If item is the first item of a new group, create new stream
        if item.start in self._group_start_to_group:
            group = self._group_start_to_group[item.start]
            self._stream_state = _StreamState(
                stream=self._get_stream(group.start, group.end),
                stream_position=group.start,
                leftover=None,
            )
            assert self._stream_state.stream is not None
            return self._stream_state.stream

        # Otherwise, we're still in same group - reuse stream created when reading 1st item
        assert (  # Assert mainly serves for mypy checks.
            self._stream_state.stream is not None
        ), "No stream found for item; first item of its range group likely not read"
        return self._stream_state.stream

    def _get_item_buffer(self, item: ItemRange) -> _ItemViewBuffer:
        """Load entire item into a memoryview-segment buffer from existing stream.

        1. Handles leftover bytes from previous reads
        2. Skips gap data from coalescing within <=max_gap_size
        3. Fetches item data from S3 stream into buffer

        Returns buffer ready for read/readinto calls.
        """

        buffer = _ItemViewBuffer()

        # Get stream from the right RangeGroup for start_pos
        stream = self._get_stream_for_item(item)
        pos = self._stream_state.stream_position  # global offset in S3 object
        leftover = self._stream_state.leftover  # leftover from previous chunk
        bytes_left = item.end - item.start

        # --- Phase 1: Handle leftover bytes from previous chunk ---
        #
        # Leftover contains bytes from the end of the previous chunk (say 8MB) that weren't consumed.
        # The leftover always ends at a chunk boundary (of 8MB parts - assume 8MB from now for explanation)
        #
        # Two cases:
        #   A) item.start within leftover: extract needed portion, possibly
        #      i) skipping a prefix (gap data from coalescing ranges within max_gap_size), and/or
        #      ii) saving a suffix (next item's data) as new leftover
        #   B) item.start beyond leftover: discard all (gap data from coalescing ranges)
        #
        # Case A visualization (item starts within leftover):
        #
        #   8MB chunks: ...====|================================|====...
        #   leftover:                     |#####################|           (length: leftover_len)
        #                                 ^                     ^
        #                                pos             leftover_end_pos   (global position in object)
        #   Slice offsets:                |gap|used|new_leftover|           (gap/new_leftover can be empty)
        #                                     ^    ^
        #                              item.start item.end                  (global position in object)
        #                       start_in_leftover  end_in_leftover          (relative to leftover)
        #   Lengths:                          |<--------------->|
        #                                       available_bytes
        #                                     |<-->|
        #                                 bytes_to_extract
        #
        if leftover:
            leftover_len = len(leftover)
            leftover_end_pos = pos + leftover_len

            if pos <= item.start < leftover_end_pos:
                # Case A: Item starts within leftover data:
                # i) if there's gap data to skip, ignore it
                start_in_leftover = item.start - pos
                # ii) if more bytes than required, save suffix as new leftover
                available_bytes = leftover_len - start_in_leftover
                bytes_to_extract = min(bytes_left, available_bytes)
                end_in_leftover = start_in_leftover + bytes_to_extract

                # Extract needed portion to buffer, and update leftover
                buffer.append_view(leftover[start_in_leftover:end_in_leftover])
                bytes_left -= bytes_to_extract
                pos = item.start + bytes_to_extract
                leftover = (  # Update 'new_leftover'
                    leftover[end_in_leftover:]
                    if end_in_leftover < leftover_len
                    else None
                )
            else:  # i.e. leftover_end_pos <= item.start:
                # Case B: Item beyond leftover: discard leftover (it was gap data)
                pos += leftover_len
                leftover = None

        # --- Phase 2: Skip gap data (from coalescing) ---
        # Current state: pos is at chunk boundary of 8MB parts after any leftover is processed.
        # When ranges are coalesced (within max_gap_size), there may be gap data to skip.
        # So we iterate stream until chunk contains item.start.
        #
        # Two cases per chunk:
        #   A) Full chunk is gap: discard entirely, continue till pos >= item.start
        #   B) Boundary chunk: current chunk contains item.start
        #
        # Case B visualization (boundary chunk):
        #
        #   8MB chunks: ...====|================================|====...
        #   Fetched chunk:     |################################|           (length: chunk_len)
        #                      ^                                ^
        #                     pos                          pos+chunk_len
        #   Slice offsets:     | gap |   used   | new_leftover  |           (gap/new_leftover can be empty)
        #                            ^          ^
        #                     item.start     item.end                       (global position in object)
        #   Lengths:           |<--->|
        #                    skip_bytes
        #
        while pos < item.start:
            try:
                chunk = memoryview(next(stream))
            except StopIteration:
                raise ValueError(
                    f"S3 stream exhausted at position {pos} before reaching item {item.start}-{item.end}"
                )

            chunk_len = len(chunk)

            if pos + chunk_len <= item.start:
                # Entire chunk before item start - skip completely
                pos += chunk_len
                continue
            else:
                # Partial Skip - slice off unwanted part first
                skip_bytes = item.start - pos
                chunk = chunk[skip_bytes:]
                pos = item.start
                chunk_len -= skip_bytes

                # Now process boundary chunk
                if chunk_len <= bytes_left:
                    # Entire chunk needed - skip slicing
                    buffer.append_view(chunk)
                    bytes_left -= chunk_len
                    pos += chunk_len
                    leftover = None
                else:
                    # Only part of chunk needed
                    buffer.append_view(chunk[:bytes_left])
                    leftover = chunk[bytes_left:]
                    pos += bytes_left
                    bytes_left = 0
                    break

        # --- Phase 3: Fetch remaining item data ---
        # Current state: pos is at chunk boundary, and pos == item.start + [bytes extracted]
        # If bytes_left > 0, we still need more data for this item.
        #
        # Two cases per chunk:
        #   A) Full chunk needed: add all used bytes to buffer, continue if bytes_left > 0
        #   B) Partial chunk: item ends mid-chunk, add used bytes to buffer and update leftover
        #
        # Case B visualization (partial chunk):
        #
        #   8MB chunks: ...====|================================|====...
        #   Fetched chunk:     |################################|           (length: chunk_len)
        #                      ^                                ^
        #                     pos                          pos+chunk_len
        #   Slice offsets:     |  used  |      new_leftover     |
        #                      ^        ^
        #                     pos      item.end                             (global position in object)
        #   Lengths:           |<------>|
        #                      bytes_left                                   (bytes_left only if item ends here)
        #
        while bytes_left > 0:
            try:
                chunk = memoryview(next(stream))
            except StopIteration:
                raise ValueError(
                    f"S3 stream exhausted at position {pos} while reading item {item.start}-{item.end}"
                )

            chunk_len = len(chunk)

            if chunk_len <= bytes_left:
                # Entire chunk needed - skip slicing
                buffer.append_view(chunk)
                bytes_left -= chunk_len
                pos += chunk_len
                leftover = None
            else:
                # Only part of chunk needed
                buffer.append_view(chunk[:bytes_left])
                leftover = chunk[bytes_left:]  # new_leftover
                pos += bytes_left
                bytes_left = 0
                break

        self._stream_state.stream_position = pos
        self._stream_state.leftover = leftover
        return buffer

    def read(self, size: Optional[int] = None) -> bytes:
        """
        Read up to size bytes from the current position.

        Supports backward seeking within the current item buffer, but forward-only
        access across DCP items (sequential item access required).

        Args:
            size (int): how many bytes to read.

        Returns:
            bytes: Bytes read from specified range.

        Raises:
            TypeError: If size is not an integer.
            ValueError: If position is outside valid DCP ranges, and if size is None or negative (full file reads not supported).
            S3Exception: An error occurred accessing S3.
        """
        if size is None:
            raise ValueError(
                f"Size cannot be None; full read not supported.\n{FALLBACK_GUIDANCE}"
            )
        if not isinstance(size, int):
            raise TypeError(f"argument should be integer or None, not {type(size)!r}")
        if size < 0:
            raise ValueError(
                f"Size cannot be negative; full read not supported.\n{FALLBACK_GUIDANCE}"
            )
        if size == 0:
            return b""

        item = self._find_item_for_range(self._position, self._position + size)

        # if item has been changed (or first item), then load new item to buffer
        if item is not self._current_item or self._current_item_buffer is None:
            self._current_item = item
            self._current_item_buffer = self._get_item_buffer(item)

        # Convert global position to item-relative offset for buffer seek
        local_pos_in_item_buffer = self._position - item.start
        self._current_item_buffer.seek(local_pos_in_item_buffer)
        data = self._current_item_buffer.read(size)

        self._position += len(data)
        return data

    def readinto(self, buf) -> int:
        """
        Read up to len(buf) bytes into a pre-allocated, writable bytes-like object buf.
        Return the number of bytes read. If no bytes are available, zero is returned.

        Args:
            buf : writable bytes-like object

        Returns:
            int : number of bytes read or zero, if no bytes available

        Raises:
            ValueError: If position is outside valid DCP ranges.
            TypeError: If buf is not writable.
            S3Exception: An error occurred accessing S3.
        """
        item = self._find_item_for_range(self._position, self._position + len(buf))

        # if item has been changed (or first item), then load new item to buffer
        if item is not self._current_item or self._current_item_buffer is None:
            self._current_item = item
            self._current_item_buffer = self._get_item_buffer(item)

        # Convert global position to item-relative offset for buffer seek
        local_pos_in_item_buffer = self._position - item.start
        self._current_item_buffer.seek(local_pos_in_item_buffer)
        bytes_read = self._current_item_buffer.readinto(buf)

        self._position += bytes_read
        return bytes_read

    def seek(self, offset: int, whence: int = SEEK_SET, /) -> int:
        """
        Change position within DCP ranges, interpreted relative to whence.

        Supports arbitrary seeking within current item buffer, but only forward
        sequential access across DCP items (cannot seek back to previous items).

        Args:
            offset (int): How many bytes to seek relative to whence.
            whence (int): One of SEEK_SET, and SEEK_CUR. SEEK_END not supported. Default: SEEK_SET.

        Returns:
            int: Current position of the stream

        Raises:
            TypeError: If whence is not SEEK_SET or SEEK_CUR.
            ValueError: If seeking to negative position or accessing previous items.
        """
        if not isinstance(offset, int):
            raise TypeError(f"integer argument expected, got {type(offset)!r}")

        if whence == SEEK_SET:
            self._position = offset
        elif whence == SEEK_CUR:
            self._position += offset
        else:
            raise ValueError("whence must be SEEK_CUR or SEEK_SET integers")

        if self._position < 0:
            raise ValueError(f"negative seek value {self._position}")

        return self._position

    def tell(self) -> int:
        """
        Returns:
            int: Current absolute position in the object.
        """
        return self._position

    def close(self) -> None:
        """
        Close the stream and release resources.
        """
        if not self._closed:
            self._closed = True
            self._stream_state.stream = None
            self._stream_state.leftover = None
            if self._current_item_buffer:
                self._current_item_buffer = None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

import os
from typing import (
    TYPE_CHECKING,
    Protocol,
    Callable,
    Optional,
    Union,
    List,
    Dict,
    runtime_checkable,
)
from .s3reader import S3Reader
from .dcp_optimized import ItemRange
from s3torchconnectorclient._mountpoint_s3_client import (
    ObjectInfo,
    GetObjectStream,
    HeadObjectResult,
)

if TYPE_CHECKING:
    from torch.distributed.checkpoint.planner import ReadItem
    from torch.distributed.checkpoint.metadata import MetadataIndex
    from torch.distributed.checkpoint.filesystem import _StorageInfo


class GetStreamCallable(Protocol):
    def __call__(
        self, start: Optional[int] = None, end: Optional[int] = None
    ) -> GetObjectStream: ...


@runtime_checkable
class S3ReaderConstructorProtocol(Protocol):
    def __call__(
        self,
        bucket: str,
        key: str,
        get_object_info: Callable[[], Union[ObjectInfo, HeadObjectResult]],
        get_stream: GetStreamCallable,
    ) -> S3Reader: ...


@runtime_checkable
class DCPS3ReaderConstructorProtocol(Protocol):
    _item_ranges_by_file: Dict[str, List[ItemRange]]

    def __call__(
        self,
        bucket: str,
        key: str,
        get_object_info: Callable[[], Union[ObjectInfo, HeadObjectResult]],
        get_stream: GetStreamCallable,
    ) -> S3Reader: ...

    def set_item_ranges_by_file(
        self,
        plan_items: "List[ReadItem]",
        storage_data: "Dict[MetadataIndex, _StorageInfo]",
        base_path: Union[str, os.PathLike],
    ) -> None: ...



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  // SPDX-License-Identifier: BSD

from .s3client_config import S3ClientConfig
from ._s3client import S3Client
from ._mock_s3client import MockS3Client

__all__ = [
    "S3ClientConfig",
    "S3Client",
    "MockS3Client",
]
